// N-Queen Problem Hopfield-Tank Neural Network
// Written by David J. Stein, Esq. See "License.html" for terms and conditions of use.
//
// This module contains the Hopfield-Tank network class.

#region Using declarations

using System;

#endregion

namespace Hopfield_Tank_Neural_Network_Namespace {

	public class HopfieldTankNetwork	{

		#region Variables and initialization code

		internal int size;
		internal double [,] V;
		internal double [,] VPrevious;
		internal double [,] u;
		internal double [,] uInitial;
		internal double E;
		internal double u0 = 0.5;
		internal double du0 = 0.05;
		internal double dt = 0.01;
		Random r;

		public HopfieldTankNetwork(int size)	{
			this.size = size;
			V = new double[size,size];
			VPrevious = new double[size,size];
			u = new double[size,size];
			uInitial = new double[size,size];
			r = new Random();
			Initialize();
		}

		#endregion

		#region Core functions

		public void Initialize() {
			// This function randomizes the network for a new solution.
			for (int i = 0; i < size; i++) {
				for (int j =0 ; j < size; j++) {
					u[i,j] = uInitial[i,j] = u0 + (r.Next(100) / 100.0) * du0 - (du0 / 2.0);
					V[i,j] = VPrevious[i,j] = 0;
				}
			}
		}

		public void Reset() {
			// This function restores the initial-state values of the current solution.
			for (int i = 0; i < size; i++) {
				for (int j =0 ; j < size; j++) {
					u[i,j] = uInitial[i,j];
					V[i,j] = VPrevious[i,j] = 0;
				}
			}
		}

		public void Analyze() {
			// This function moves the neural network one step toward a solution.

			// update V values
			for (int i = 0; i < size; i++) {
				for (int j = 0; j < size; j++) {
					VPrevious[i,j] = V[i,j];
					V[i,j] = 0.5 * (1.0 + Math.Tanh(u[i,j] / u0));
				}
			}

			// update u values
			for (int i = 0; i < size; i++) {
				for (int j = 0; j < size; j++) {

					// calculate first term (vertical counting)
					double ASum = -1.0;
					for (int k = 0; k < size; k++)
						ASum += V[i,k];

					// calculate second term (horizontal counting)
					double BSum = -1.0;
					for (int k = 0; k < size; k++)
						BSum += V[k,j];

					// calculate third sum (rising diagonal)
					double CSum = 0.0;
					for (int k = Math.Max(-i, -j); k <= Math.Min(size - 1 - i, size - 1 - j); k++) {
						if (k != 0) {
							CSum += V[i+k,j+k];
						}
					}

					// calculate fourth sum (falling diagonal)
					double DSum = 0.0;
					for (int k = Math.Max(-i, j - size - 1); k <= Math.Min(size - 1 - i, j - 1); k++) {
						if ((k != 0) && (i + k < size) && (j - k >= 0) && (j - k < size)) {
							DSum += V[i+k,j-k];
						}
					}
					// calculate du/dt and add to u
					double dudt = -(ASum + BSum + CSum + DSum);
					u[i,j] += dudt * dt;
				}
			}

			// calculate E

			// calculate first sum
			double EASum = 0.0;
			for (int i = 0; i < size; i++) {
				double IntermediateEASum = -1.0;
				for (int k = 0; k < size; k++) {
					IntermediateEASum += V[i,k];
				}
				EASum += Math.Pow(IntermediateEASum, 2);
			}

			// calculate second sum
			double EBSum = 0.0;
			for (int j = 0; j < size; j++) {
				double IntermediateEBSum = -1.0;
				for (int k = 0; k < size; k++) {
					IntermediateEBSum += V[k,j];
				}
				EBSum += Math.Pow(IntermediateEBSum, 2);
			}

			// calculate third sum
			double ECSum = 0.0;
			for (int i = 0; i < size; i++) {
				for (int j = 0; j < size; j++) {
					for (int k = 1; k < size; k++) {
						if ((i + k < size) && (j + k < size))
							ECSum += V[i+k,j+k] * V[i,j];
					}
				}
			}

			// calculate fourth sum
			double EDSum = 0.0;
			for (int i = 0; i < size; i++) {
				for (int j = 0; j < size; j++) {
					for (int k = 1; k < size; k++) {
						if ((i + k < size) && (j - k >= 0))
							EDSum += V[i+k,j-k] * V[i,j];
					}
				}
			}

			// calculate E
			E = 0.5 * (EASum + EBSum + ECSum + EDSum);
		}

		#endregion

	}
}
